/*
 * Copyright (c) 1996, 2002 Sandro Sigala.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <math.h>
#include <stdarg.h>
#include <stdio.h>

#include "gamewidget.h"
#include "images.h"

static GameWidget *widgetptr = 0;

static void showerror(const char *fmt, ...)
{
	va_list ap;
	char buf[1024];
	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);
	FXMessageBox::error(widgetptr, MBOX_OK, "Error", buf);
}

static void showanimation(Animation a)
{
	widgetptr->showAnimation(a);
}

static void playsound(Sound snd)
{
	widgetptr->playSound(snd);
}

FXDEFMAP(GameWidget) GameWidgetMap[] = {
  FXMAPFUNC(SEL_PAINT, 0, GameWidget::onPaint),
};

FXIMPLEMENT(GameWidget, FXCanvas, GameWidgetMap, ARRAYNUMBER(GameWidgetMap))

GameWidget::GameWidget()
{
}

GameWidget::GameWidget(FXComposite *p, FXObject *tgt, FXSelector sel, FXuint opts,
		       FXint x, FXint y, FXint w, FXint h)
	: FXCanvas(p, tgt, sel, opts, x, y, w, h)
{
	zoom = 1;
	computeGameSize();

       	dbuf = new FXBMPImage(getApp(), 0, IMAGE_SHMI|IMAGE_SHMP, getWidth(), getHeight());
	dbuf->create();

	for (int i = 0; i < num_sprites; ++i) {
		sprites[i] = new FXBMPImage(getApp(), 0, IMAGE_KEEP|IMAGE_OWNED|IMAGE_SHMI|IMAGE_SHMP,
					    sprite_width, sprite_height);
		sprites[i]->create();
	}
	for (int i = 0; i < num_masks; ++i) {
		masks[i] = new FXBMPImage(getApp(), 0, IMAGE_KEEP|IMAGE_OWNED|IMAGE_SHMI|IMAGE_SHMP,
					  sprite_width, sprite_height);
		masks[i]->create();
	}
	pausebmp = new FXBMPImage(getApp(), img_pause, IMAGE_KEEP|IMAGE_SHMI|IMAGE_SHMP);
	pausebmp->create();

	widgetptr = this;
	engine.setErrorFunc(showerror);
	engine.setSoundFunc(playsound);
	engine.setAnimationFunc(showanimation);
}

GameWidget::~GameWidget()
{
	for (int i = 0; i < num_sprites; ++i)
		delete sprites[i];
	for (int i = 0; i < num_masks; ++i)
		delete masks[i];
	delete pausebmp;
	
	delete dbuf;
}

void GameWidget::showAnimation(Animation a)
{
	anim = a;
	anim_status = 0;
}

void GameWidget::computeGameSize()
{
 	xb = (getWidth() % (sprite_width*zoom)) / 2;
	gwidth = getWidth() / (sprite_width*zoom);
	yb = (getHeight() % (sprite_height*zoom)) / 2;
	gheight = getHeight() / (sprite_height*zoom);
	scr_dwidth = gwidth / 2;
	scr_dheight = gheight / 2;
	if (scr_dwidth == 0)
		++scr_dwidth;
	if (scr_dheight == 0)
		++scr_dheight;
}

void GameWidget::loadLevel(const FXString& episodepath, const FXString& spritespath, int levelno, bool resetscore)
{
	loadSprites(spritespath);
	FXString s;
	s = episodepath;
	s += PATHSEPSTRING "map";
 	s += (char)(levelno + '0');
 	s += ".fxr";
	engine.loadData(s.text(), resetscore);

	anim = Animation_None;
	anim_status = 0;
}

void GameWidget::setZoomValue(int value)
{
	if (zoom != value) {
		getApp()->beginWaitCursor();
		zoom = value;
		computeGameSize();
	 	for (int i = 0; i < num_sprites; ++i) {
	 		sprites[i]->restore();
			sprites[i]->scale(sprite_width*zoom, sprite_height*zoom);
		}
	 	for (int i = 0; i < num_masks; ++i) {
	 		masks[i]->restore();
			masks[i]->scale(sprite_width*zoom, sprite_height*zoom);
		}
		getApp()->endWaitCursor();
	}
}

void GameWidget::loadSprites(const FXString& spritespath)
{
	FXString filename = spritespath + PATHSEPSTRING "sprites.bmp";

	getApp()->beginWaitCursor();

	if (zoom != 1) {
	 	for (int i = 0; i < num_sprites; ++i)
			sprites[i]->resize(sprite_width, sprite_height);
	 	for (int i = 0; i < num_masks; ++i)
			masks[i]->resize(sprite_width, sprite_height);
	}

	FXBMPImage *img = new FXBMPImage(getApp(), 0);

	// Load it
  	FXFileStream stream;
  	if(!stream.open(filename, FXStreamLoad)) {
    		FXMessageBox::error(this, MBOX_OK, "Error", "Cannot load %s", filename.text());
  		exit(1);
     	}

	img->loadPixels(stream);
	stream.close();
	img->create();

	if (img->getWidth() != 8*sprite_width || img->getHeight() != 6*sprite_height) {
    		FXMessageBox::error(this, MBOX_OK, "Error", "The sprite bitmap %s must be 256x192", filename.text());
  		exit(1);
     	}

	// Load the sprites
	for (int y = 0; y < 6; ++y)
		for (int x = 0; x < 8; ++x) {
			int i = y*8 + x;
			FXDCWindow dc(sprites[i]);
		   	dc.setFunction(BLT_SRC);
			dc.drawImage(img, -x*sprite_width, -y*sprite_height);
			if (i == Obj_Spirit || i == Obj_Spirit2 ||
   			    i == Obj_Monster || i == Obj_Monster2) {
				FXDCWindow dc2(masks[Obj_Mask_Spirit + i - Obj_Spirit]);
			   	dc2.setFunction(BLT_SRC);
				dc2.drawImage(img, -x*sprite_width, -y*sprite_height);
   			}
   		}

	delete img;

	// Generate the masks for the monsters
 	for (int i = 0; i < num_masks; ++i) {
 		masks[i]->restore();
		for (int y = 0; y < masks[i]->getHeight(); ++y)
			for (int x = 0; x < masks[i]->getWidth(); ++x) {
				FXColor pix = masks[i]->getPixel(x, y);
				if (!FXREDVAL(pix) && !FXGREENVAL(pix) && !FXBLUEVAL(pix))
					masks[i]->setPixel(x, y, FXRGB(255, 255, 255));
				else
					masks[i]->setPixel(x, y, FXRGB(0, 0, 0));
			}
		masks[i]->render();
	}

	if (zoom != 1) {
	 	for (int i = 0; i < num_sprites; ++i) {
	 		sprites[i]->restore();
			sprites[i]->scale(sprite_width*zoom, sprite_height*zoom);
		}
	 	for (int i = 0; i < num_masks; ++i) {
	 		masks[i]->restore();
			masks[i]->scale(sprite_width*zoom, sprite_height*zoom);
		}
	}

	getApp()->endWaitCursor();
}

void GameWidget::create()
{
	FXCanvas::create();
}

FXint GameWidget::getDefaultWidth()
{
	return gwidth;
}

FXint GameWidget::getDefaultHeight()
{
	return gheight;
}

void GameWidget::drawBoard()
{
	bool resized = false;
	if (last_width != getWidth() || last_height != getHeight()) {
		resized = true;
		dbuf->resize(getWidth(), getHeight());
		computeGameSize();
		last_width = getWidth();
		last_height = getHeight();
	}

	FXDCWindow dc(dbuf);

	if (resized || anim == Animation_Transport) {
		dc.setForeground(FXRGB(0, 0, 0));
		dc.fillRectangle(0, 0, getWidth(), getHeight());
	}

	if (engine.status != Status_Idle) {
		if (anim == Animation_Transport) {
			const int maxanim = 12;
			int spacingx = getWidth()/(maxanim*2);
			int spacingy = getHeight()/(maxanim*2);
			int spacing = FXMIN(spacingx, spacingy);
			if (anim_status > maxanim)
				anim = Animation_None;
			else if (!engine.paused) {
				for (int i = anim_status; i > 0; --i) {
     					FXColor c;
					if (i % 4 == 0)
    						c = FXRGB(0, 0, 255);
					else if (i % 4 == 1)
    						c = FXRGB(0, 255, 0);
					else if (i % 4 == 2)
    						c = FXRGB(255, 255, 0);
					else
    						c = FXRGB(0, 0, 0);
					dc.setForeground(c);
					dc.fillRectangle(getWidth()/2-i*spacing, getHeight()/2-i*spacing,
							 i*spacing*2, i*spacing*2);
				}
    			}
			++anim_status;
		} else if (anim == Animation_Died && anim_status > 0) {
			int rx = scr_dwidth*sprite_width*zoom + xb;
     			int ry = scr_dheight*sprite_height*zoom + yb;
			if (anim_status == 1) {
				dc.setForeground(FXRGB(0, 0, 0));
				dc.fillRectangle(rx, ry, sprite_width*zoom, sprite_height*zoom);
    			} else if (anim_status > 50)
				anim = Animation_None;
			++anim_status;
  			for (int i = 0; i < 10; ++i) {
      				FXColor c = FXRGB(rand()%10, rand()%256, rand()%10);
				dc.setForeground(c);
				float r = (float)rand() / RAND_MAX * (sprite_width*zoom*(anim_status<25?anim_status:25)/25)/2;
				float a = (float)rand() / RAND_MAX * PI*2;
			  	dc.fillRectangle(rx + sprite_width*zoom/2 - 1 + (int)(cos(a)*r),
					         ry + sprite_height*zoom/2 - 1 + (int)(sin(a)*r), zoom, zoom);
			}
		} else { // anim == Animation_None || (anim == Animation_Died && anim_status == 0)
			for (int y = 0; y < gheight; ++y) {
				for (int x = 0; x < gwidth; ++x) {
					SpriteObj obj;
					int mx = x + engine.xpos - scr_dwidth;
					int my = y + engine.ypos - scr_dheight;
					obj = engine.xy(mx, my);
     					if (engine.status == Status_Died && obj == Obj_Repton)
						continue;
					switch (obj) {
					case Obj_Broken_Egg1:
					case Obj_Broken_Egg2:
					case Obj_Broken_Egg3:
						obj = Obj_Broken_Egg;
						break;
					case Obj_Falling_Egg:
						obj = Obj_Egg;
						break;
					case Obj_Repton:
						obj = engine.sprite;
						break;
					default:;	// Avoid warning
					}

				   	dc.setFunction(BLT_SRC);
					dc.drawImage(sprites[obj], x*sprite_width*zoom + xb, y*sprite_height*zoom + yb);

					Monster *m;
					m = engine.findMonster(mx, my);
					if (m != 0) {
					   	dc.setFunction(BLT_SRC_AND_DST);
						dc.drawImage(masks[m->sprite - Obj_Spirit + Obj_Mask_Spirit],
     							     x*sprite_width*zoom + xb, y*sprite_height*zoom + yb);
					  		dc.setFunction(BLT_SRC_OR_DST);
						dc.drawImage(sprites[m->sprite], x*sprite_width*zoom + xb, y*sprite_height*zoom + yb);
     					}
				}
			}
			if (anim == Animation_Died)
				++anim_status;
		}

		if (engine.status == Status_Playing && engine.paused) {
			dc.setFunction(BLT_SRC);
			dc.drawImage(pausebmp, (getWidth() - pausebmp->getWidth()) / 2,
				     (getHeight() - pausebmp->getHeight()) / 2);
		}
	}
}

long GameWidget::onPaint(FXObject *, FXSelector, void *ptr)
{
	drawBoard();

 	FXEvent *ev = (FXEvent *)ptr;
	FXDCWindow dc(this, ev);
  	dc.drawImage(dbuf, 0, 0);
	return 1;
}

