/*
 * Copyright (c) 1996, 2002 Sandro Sigala.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <ctype.h>
#include <dirent.h>
#include <stdio.h>
#include <algorithm>

#include <fxkeys.h>

#include "mainwindow.h"
#include "gamewidget.h"
#include "legendwindow.h"
#include "mapwindow.h"
#include "paths.h"
#include "version.h"

#include "fxrepton.xpm"
#include "fxrepton_mini.xpm"

FXDEFMAP(MainWindow) MainWindowMap[] = {
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_RESTART,   MainWindow::onCmdRestart),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_RESTARTEP, MainWindow::onCmdRestartEpisode),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_PASSWORD,  MainWindow::onCmdPassword),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_PAUSE,     MainWindow::onCmdPause),
  FXMAPFUNC(SEL_UPDATE,    MainWindow::ID_PAUSE,     MainWindow::onUpdPause),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_ICONIFY,   MainWindow::onCmdIconify),
  FXMAPFUNC(SEL_SIGNAL,    MainWindow::ID_QUIT,      MainWindow::onCmdQuit),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_QUIT,      MainWindow::onCmdQuit),
  FXMAPFUNCS(SEL_COMMAND,  MainWindow::ID_EPISODE,   MainWindow::ID_EPISODELAST, MainWindow::onCmdEpisode),
  FXMAPFUNCS(SEL_UPDATE,   MainWindow::ID_EPISODE,   MainWindow::ID_EPISODELAST, MainWindow::onUpdEpisode),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_SOUND,     MainWindow::onCmdSound),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_ZOOM,      MainWindow::onCmdZoom),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_LEGEND,    MainWindow::onCmdLegend),
  FXMAPFUNC(SEL_UPDATE,    MainWindow::ID_LEGEND,    MainWindow::onUpdLegend),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_MAP,       MainWindow::onCmdMap),
  FXMAPFUNC(SEL_UPDATE,    MainWindow::ID_MAP,       MainWindow::onUpdMap),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_ABOUT,     MainWindow::onCmdAbout),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_UP,  	     MainWindow::onCmdUp),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_DOWN,      MainWindow::onCmdDown),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_LEFT,      MainWindow::onCmdLeft),
  FXMAPFUNC(SEL_COMMAND,   MainWindow::ID_RIGHT,     MainWindow::onCmdRight),
  FXMAPFUNC(SEL_TIMEOUT,   MainWindow::ID_TIMER,     MainWindow::onTimer),
};

FXIMPLEMENT(MainWindow, FXMainWindow, MainWindowMap, ARRAYNUMBER(MainWindowMap))

MainWindow::MainWindow()
{
}

MainWindow::MainWindow(FXApp *app)
    : FXMainWindow(app, "FX-Repton", 0, 0, DECOR_TITLE|DECOR_MINIMIZE|DECOR_RESIZE|DECOR_CLOSE|DECOR_BORDER|DECOR_MENU, 0, 0, 360, 360)
{
	bigicon = new FXXPMIcon(getApp(), fxrepton_xpm);
  	smallicon = new FXXPMIcon(getApp(), fxrepton_mini_xpm);
  	setIcon(bigicon);
    	setMiniIcon(smallicon);

	legend = 0;
	map = 0;

	loadEpisodeList(PATH_EPISODES);
	loadEpisodeList(FXFile::getHomeDirectory() + PATHSEPSTRING PATH_USER_EPISODES);
	if (episodes.size() < 2) {
		fprintf(stderr, "Cannot load episodes!\n");
		exit(1);
	}

       	FXMenubar *menubar;

        FXToolbarShell *dragshell1 = new FXToolbarShell(this, FRAME_RAISED);
        menubar = new FXMenubar(this, dragshell1, FRAME_RAISED|LAYOUT_SIDE_TOP|LAYOUT_FILL_X);
        new FXToolbarGrip(menubar, menubar, FXMenubar::ID_TOOLBARGRIP);

       	FXStatusbar *statusbar;
	statusbar = new FXStatusbar(this, LAYOUT_SIDE_BOTTOM|LAYOUT_FILL_X|STATUSBAR_WITH_DRAGCORNER|FRAME_RAISED);

        filemenu = new FXMenuPane(this);
        new FXMenuCommand(filemenu, "&Restart level\tF2\tRestart current level.", 0, this, ID_RESTART);
        new FXMenuCommand(filemenu, "&Restart episode\t\tRestart current episode.", 0, this, ID_RESTARTEP);
  	new FXMenuSeparator(filemenu);
        new FXMenuCommand(filemenu, "&Enter password...\t\tStart a specific level.", 0, this, ID_PASSWORD);
  	new FXMenuSeparator(filemenu);
        new FXMenuCommand(filemenu, "&Pause\tP\tPause the game.", 0, this, ID_PAUSE);
	new FXMenuSeparator(filemenu);
        new FXMenuCommand(filemenu, "&Quit\tAlt-X\tClose this program.", 0, this, ID_QUIT);
        new FXMenuTitle(menubar, "&Game", 0, filemenu);

        episodemenu = new FXMenuPane(this);
	FXMenuPane *lastPane = 0;
	FXString lastPackage;
	typedef vector<Episode>::const_iterator I;
	for (I i = episodes.begin(); i != episodes.end(); ++i) {
  		Episode ep = *i;
		if (!lastPane || lastPackage != ep.package) {
			// remove "[n] " from "[n] PackageName"
			FXString package = ep.package;
  			if (package.findf(']') != -1)
				package = package.after(']').trim();
			FXMenuCascade *cascade = new FXMenuCascade(episodemenu, package);
        		lastPane = new FXMenuPane(this);
        		lastPackage = ep.package;
			cascade->setMenu(lastPane);
		}
		// remove "[n] " from "[n] EpisodeName"
		FXString name = ep.name;
  		if (name.findf(']') != -1)
			name = name.after(']').trim();
		new FXMenuCommand(lastPane, name, 0, this, ep.id);
	}
        new FXMenuTitle(menubar, "&Episode", 0, episodemenu);

        optionsmenu = new FXMenuPane(this);
        new FXMenuCommand(optionsmenu, "&Sound\t\tEnable/disable sound.", 0, &sound_target, FXDataTarget::ID_VALUE);
  	new FXMenuSeparator(optionsmenu);
        new FXMenuCommand(optionsmenu, "&Zoom x1\t\tSet the zoom factor to 1x.", 0, &zoom_target, FXDataTarget::ID_OPTION+0);
        new FXMenuCommand(optionsmenu, "&Zoom x2\t\tSet the zoom factor to 2x.", 0, &zoom_target, FXDataTarget::ID_OPTION+1);
        new FXMenuCommand(optionsmenu, "&Zoom x3\t\tSet the zoom factor to 3x.", 0, &zoom_target, FXDataTarget::ID_OPTION+2);
        new FXMenuTitle(menubar, "&Options", 0, optionsmenu);

        helpmenu = new FXMenuPane(this);
        new FXMenuCommand(helpmenu, "&Show legend\t\tShow game legend.", 0, this, ID_LEGEND);
        new FXMenuCommand(helpmenu, "&Show map\t\tShow game map.", 0, this, ID_MAP);
  	new FXMenuSeparator(helpmenu);
        new FXMenuCommand(helpmenu, "&About...\t\tShow information about this program.", 0, this, ID_ABOUT);
        new FXMenuTitle(menubar, "&?", 0, helpmenu);

       	game = new GameWidget(this, this, ID_GAME, LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y,
        	0,0,default_game_width,default_game_height);

  	level_label = new FXLabel(statusbar, "   ", 0, FRAME_SUNKEN|LAYOUT_RIGHT|LAYOUT_CENTER_Y);
  	score_label = new FXLabel(statusbar, " 0 ", 0, FRAME_SUNKEN|LAYOUT_RIGHT|LAYOUT_CENTER_Y);
 	new FXLabel(statusbar," Score:",0,LAYOUT_RIGHT|LAYOUT_CENTER_Y);
  	diamonds_label = new FXLabel(statusbar, " 0 ", 0, FRAME_SUNKEN|LAYOUT_RIGHT|LAYOUT_CENTER_Y);
 	new FXLabel(statusbar," Diamonds:",0,LAYOUT_RIGHT|LAYOUT_CENTER_Y);
  	time_label = new FXLabel(statusbar, " 0 ", 0, FRAME_SUNKEN|LAYOUT_RIGHT|LAYOUT_CENTER_Y);
 	new FXLabel(statusbar," Time:",0,LAYOUT_RIGHT|LAYOUT_CENTER_Y);

	getApp()->reg().read();
	game->engine.sounds = getApp()->reg().readUnsignedEntry("Settings", "Sounds", 1);

	sound_target.connect(game->engine.sounds);
  	sound_target.setTarget(this);
  	sound_target.setSelector(ID_SOUND);

	zoom = 0;
  	zoom_target.connect(zoom);
	zoom_target.setTarget(this);
  	zoom_target.setSelector(ID_ZOOM);

	FXAccelTable *t = getAccelTable();
	if (t) {
    		t->addAccel(KEY_Escape, this, MKUINT(ID_ICONIFY, SEL_COMMAND));
    		t->addAccel(KEY_Up,     this, MKUINT(ID_UP,      SEL_COMMAND));
    		t->addAccel(KEY_Down,   this, MKUINT(ID_DOWN,    SEL_COMMAND));
    		t->addAccel(KEY_Left,   this, MKUINT(ID_LEFT,    SEL_COMMAND));
    		t->addAccel(KEY_Right,  this, MKUINT(ID_RIGHT,   SEL_COMMAND));
    	}

	lastkeytime = 0;

	// Default episode: Prelude (colors enhanced)
	episode = episodes[1];
	episode.level = 1;

	timerhandle = getApp()->addTimeout(100, this, ID_TIMER);
}

MainWindow::~MainWindow()
{
	delete filemenu;
	delete episodemenu;
	delete optionsmenu;
	delete helpmenu;
	getApp()->removeTimeout(timerhandle);
	if (legend) delete legend;
	if (map) delete map;

	getApp()->reg().writeUnsignedEntry("Settings", "Sounds", game->engine.sounds);
	getApp()->reg().write();
}

void MainWindow::create()
{
  	FXMainWindow::create();
}

bool operator< (const Episode& e1, const Episode& e2)
{
	if (e1.package[0] == '[' && e2.package[0] != '[')
		return true;
	if (e1.package[0] != '[' && e2.package[0] == '[')
		return false;
	return (e1.package < e2.package) || (e1.package == e2.package && e1.name < e2.name);
}

void MainWindow::loadEpisodeList(const FXString& epath)
{
	DIR *dir = opendir(epath.text());
 	struct dirent *entry;

	if (!dir) {
		fprintf(stderr, "Warning: cannot load episodes from %s\n", epath.text());
		return;
	}

	static int id = ID_EPISODE;
	while ((entry = readdir(dir)) != NULL) {
		FXString path = epath;
		path += PATHSEP;
  		path += entry->d_name;
		FXString fname = path + PATHSEPSTRING "info.fxr";

		FILE *fin = fopen(fname.text(), "r");
		if (fin) {
     			char buf[1024];
 			FXString package, epname;
     			int nolevels;
     			int broken = 0;
     			fgets(buf, 1024, fin); package = FXString(buf).trim();
     			fgets(buf, 1024, fin); epname = FXString(buf).trim();
     			fgets(buf, 1024, fin); sscanf(buf, "%d %d", &nolevels, &broken);
		     	fclose(fin);
			if (!broken) {
 				Episode ep(package, epname, path, nolevels, id++);
				episodes.push_back(ep);
			}
    		}
	}
	closedir(dir);

	sort(episodes.begin(), episodes.end());
}

// Lazy user protection.
/* TODO (#5#): Works only on ASCII machines */
static FXString rot13(const FXString& str)
{
	FXString s = str;
 	for (int i = 0; i < s.length(); ++i) {
 		unsigned char c = s[i];
		if (c >= 'a' && c <= 'z') {
 			c = c + 13;
 			if (c > 'z')
 				c = 'a' + c - 'z' - 1;
			s[i] = c;
		}
	}
	return s;
}

void MainWindow::startLevel(const Episode& e, int levelno, bool resetscore)
{
	if (episode.spritespath != e.spritespath) {
 		if (legend) legend->updateSprites(e.spritespath);
		if (map) map->updateSprites(e.spritespath);
	}

	episode = e;
	episode.level = levelno;
	game->loadLevel(episode.path, episode.spritespath, episode.level, resetscore);
	game->engine.status = Status_Playing;
	if (map) map->updateMap();

	FXString pwd = game->engine.password;
	pwd = rot13(pwd);
	FXString s;
	s = "FX-Repton [password: ";
 	s += pwd;
  	s += "]";
	setTitle(s);
}

long MainWindow::onCmdRestart(FXObject *, FXSelector, void *)
{
	if (game->engine.status != Status_Playing ||
 		FXMessageBox::question(this, MBOX_YES_NO, "FX-Repton", "Are you sure to restart the level?") == MBOX_CLICKED_YES)
		startLevel(episode, episode.level, true);
	return 1;
}

long MainWindow::onCmdRestartEpisode(FXObject *, FXSelector, void *)
{
	if (game->engine.status != Status_Playing ||
 		FXMessageBox::question(this, MBOX_YES_NO, "FX-Repton", "Are you sure to restart the episode?") == MBOX_CLICKED_YES)
		startLevel(episode, 1, true);
	return 1;
}

long MainWindow::onCmdPassword(FXObject *, FXSelector, void *)
{
	FXString str;
 	if (!FXInputDialog::getString(str, this, "FX-Repton", "Enter password:"))
  		return 1;

	for (int i = 1; i <= episode.nolevels; ++i) {
		FXString fname = episode.path + PATHSEPSTRING "map";
  		fname += (char)(i + '0');
		fname += ".fxr";

		FILE *fin = fopen(fname.text(), "r");
		if (fin) {
			char buf[1024];
			FXString pwd;
			fgets(buf, 1024, fin); // Ignore map geometry
			fgets(buf, 1024, fin); // Get password
			pwd = FXString(buf).trim();
		     	fclose(fin);
        		pwd = rot13(pwd);
        		if (pwd.lower() == str.trim().lower()) {
				FXString msg;
				msg.format("Password correct!  Press OK to play level %d!", i);
	  			FXMessageBox::information(this, MBOX_OK, "FX-Repton", msg.text());
				startLevel(episode, i, true);
    				return 1;
			}
    		}
	}

	FXMessageBox::information(this, MBOX_OK, "FX-Repton", "Password not recognized!");

	return 1;
}

long MainWindow::onCmdPause(FXObject *, FXSelector, void *)
{
	if (game->engine.status == Status_Playing)
		game->engine.paused = !game->engine.paused;
	return 1;
}

long MainWindow::onUpdPause(FXObject *sender, FXSelector, void *)
{
	FXint checked = game->engine.status == Status_Playing && game->engine.paused;
	sender->handle(this, MKUINT(ID_SETINTVALUE, SEL_COMMAND), (void*)&checked);
  	FXuint enabled = (game->engine.status == Status_Playing) ? ID_ENABLE : ID_DISABLE;
   	sender->handle(this, MKUINT(enabled, SEL_COMMAND), 0);
	return 1;
}

long MainWindow::onCmdIconify(FXObject *, FXSelector, void *)
{
	iconify();
	return 1;
}

long MainWindow::onCmdQuit(FXObject *, FXSelector, void *)
{
	getApp()->exit(0);
 	return 1;
}

long MainWindow::onCmdEpisode(FXObject *, FXSelector sel, void *)
{
	FXint x, y;
 	FXuint buttons;
	getCursorPosition(x, y, buttons);
	if (buttons & SHIFTMASK || game->engine.status != Status_Playing ||
 		FXMessageBox::question(this, MBOX_YES_NO, "FX-Repton", "Are you sure to start another episode?") == MBOX_CLICKED_YES) {
		typedef vector<Episode>::const_iterator I;
		for (I i = episodes.begin(); i != episodes.end(); ++i) {
  			Episode ep = *i;
  			if (ep.id == SELID(sel)) {
  				if (buttons & SHIFTMASK) {
  					episode.spritespath = ep.spritespath;
					game->loadSprites(ep.spritespath);
			  		if (legend) legend->updateSprites(ep.spritespath);
					if (map) {
    						map->updateSprites(ep.spritespath);
	        				map->updateMap();
    					}
    				} else
	            			startLevel(ep, 1, true);
				break;
			}
		}
	}
	return 1;
}

long MainWindow::onUpdEpisode(FXObject *sender, FXSelector sel, void *)
{
	typedef vector<Episode>::const_iterator I;
	for (I i = episodes.begin(); i != episodes.end(); ++i) {
  		Episode ep = *i;
  		if (ep.id == SELID(sel)) {
			FXint checked = episode.id == ep.id;
		    	sender->handle(this, MKUINT(ID_SETINTVALUE, SEL_COMMAND), (void*)&checked);
			break;
		}
	}
	return 1;
}

long MainWindow::onCmdSound(FXObject *, FXSelector, void *)
{
	game->engine.sounds = !game->engine.sounds;
	return 1;
}

long MainWindow::onCmdZoom(FXObject *, FXSelector, void *)
{
	int value = game->getZoomValue();
	if (value != (zoom+1)) {
		resize(getWidth()/value*(zoom+1), getHeight()/value*(zoom+1));
		game->setZoomValue(zoom+1);
	}
	return 1;
}

long MainWindow::onCmdLegend(FXObject *, FXSelector, void *)
{
	if (!legend) {
		legend = new LegendWindow(getApp(), "FX-Repton Legend", episode.spritespath, getIcon(), getMiniIcon());
		legend->create();
	}
	if (!legend->shown())
      		legend->show(PLACEMENT_OWNER);
	else
      		legend->hide();
	return 1;
}

long MainWindow::onUpdLegend(FXObject *sender, FXSelector, void *)
{
	FXint checked = legend && legend->shown();
	sender->handle(this, MKUINT(ID_SETINTVALUE, SEL_COMMAND), (void*)&checked);
  	FXuint enabled = game->engine.isDataLoaded() ? ID_ENABLE : ID_DISABLE;
   	sender->handle(this, MKUINT(enabled, SEL_COMMAND), 0);
	return 1;
}

long MainWindow::onCmdMap(FXObject *, FXSelector, void *)
{
	if (!map) {
		map = new MapWindow(getApp(), "FX-Repton Map", &game->engine, episode.spritespath, getIcon(), getMiniIcon());
		map->create();
	}
	if (!map->shown()) {
      		map->show(PLACEMENT_OWNER);	
        	map->updateMap();
	} else
      		map->hide();
	return 1;
}

long MainWindow::onUpdMap(FXObject *sender, FXSelector, void *)
{
	FXint checked = map && map->shown();
	sender->handle(this, MKUINT(ID_SETINTVALUE, SEL_COMMAND), (void*)&checked);
  	FXuint enabled = game->engine.isDataLoaded() ? ID_ENABLE : ID_DISABLE;
   	sender->handle(this, MKUINT(enabled, SEL_COMMAND), 0);
	return 1;
}

long MainWindow::onCmdAbout(FXObject *, FXSelector, void *)
{
	FXDialogBox about(this, "About FX-Repton", DECOR_TITLE|DECOR_BORDER,
 		0,0,0,0, 0,0,0,0, 0,0);

  	FXVerticalFrame *info2 =
        	new FXVerticalFrame(&about, LAYOUT_SIDE_LEFT|LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 10,10,10,10, 0,0);
  	FXHorizontalFrame *cont =
        	new FXHorizontalFrame(info2, LAYOUT_SIDE_RIGHT|LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 0,0,0,0, 0,0);
  	FXHorizontalFrame *infoandicon =
        	new FXHorizontalFrame(cont, LAYOUT_SIDE_RIGHT|LAYOUT_FILL_X|LAYOUT_FILL_Y,
       	 		0,0,0,0, 0,0,0,0, 0,0);
	new FXLabel(infoandicon, FXString::null, bigicon, LAYOUT_FILL_X,
		0,0,0,0, 0,0,0,0);
  	FXVerticalFrame *info =
   		new FXVerticalFrame(infoandicon, LAYOUT_SIDE_RIGHT|LAYOUT_FILL_X|LAYOUT_FILL_Y,
   			0,0,0,0, 10,10,10,10, 0,0);
    	new FXLabel(info, "FX-Repton " VERSION "\n", 0, JUSTIFY_LEFT|LAYOUT_FILL_X);
      	new FXHorizontalSeparator(info, SEPARATOR_GROOVE|LAYOUT_FILL_X);
        new FXLabel(info,
        	"\nCopyright "
#ifdef WIN32
        	""
#else
         "(C)"
#endif
         " 2002 Sandro Sigala <sandro@sigala.it>.\n\n"
        	"Latest version can be found at http://www.sigala.it/sandro.",
      		0, JUSTIFY_LEFT|LAYOUT_FILL_X|LAYOUT_FILL_Y);
        FXButton *button = new FXButton(cont, "OK", 0, &about,
	        FXDialogBox::ID_ACCEPT, BUTTON_INITIAL|BUTTON_DEFAULT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT,
         	0,0,0,0, 32,32,5,5);

  	FILE *fp;
      	if ((fp = fopen(PATH_DATA_DIR PATHSEPSTRING "Copying", "r")) != 0) {
  	      	FXuint size, n;
  	      	FXchar *text;
  	      	fseek(fp, 0, SEEK_END);
    	      	size = ftell(fp);
  	      	fseek(fp, 0, SEEK_SET);
      	      	FXCALLOC(&text,FXchar,size+1);
      	      	n = fread(text, 1, size, fp);
      	      	fclose(fp);
        	new FXLabel(info2, "License:", 0, LAYOUT_SIDE_TOP|JUSTIFY_LEFT|LAYOUT_FILL_X|LAYOUT_FILL_Y);
        	FXVerticalFrame *textf =
        		new FXVerticalFrame(info2, FRAME_SUNKEN|LAYOUT_SIDE_TOP|LAYOUT_FIX_HEIGHT|LAYOUT_FILL_X,
       	 			0,0,0,200, 0,0,0,0, 0,0);
                FXText *textv = new FXText(textf, 0, 0, LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y);
      	      	textv->setText(text, n);
                textv->setEditable(false);
      	      	FXFREE(&text);
        }

        new FXLabel(info2,
        	"\nThis software uses the FOX Toolkit Library (http://www.fox-toolkit.org).",
      		0, JUSTIFY_LEFT|LAYOUT_FILL_X|LAYOUT_FILL_Y);

        button->setFocus();
      	about.execute(PLACEMENT_OWNER);

	return 1;
}

long MainWindow::onCmdUp(FXObject *, FXSelector, void *ptr)
{
	if (game->engine.status == Status_Playing) {
  		FXEvent *event = (FXEvent *)ptr;
		if (event->time - lastkeytime >= game_keytime) {
			game->engine.goUp();
			game->update();
			lastkeytime = event->time;
		}
	}
	return 1;
}

long MainWindow::onCmdDown(FXObject *, FXSelector, void *ptr)
{
	if (game->engine.status == Status_Playing) {
  		FXEvent *event = (FXEvent *)ptr;
		if (event->time - lastkeytime >= game_keytime) {
			game->engine.goDown();
			game->update();
			lastkeytime = event->time;
		}
	}
	return 1;
}

long MainWindow::onCmdLeft(FXObject *, FXSelector, void *ptr)
{
	if (game->engine.status == Status_Playing) {
  		FXEvent *event = (FXEvent *)ptr;
		if (event->time - lastkeytime >= game_keytime) {
			game->engine.goLeft();
			game->update();
			lastkeytime = event->time;
		}
 	}
	return 1;
}

long MainWindow::onCmdRight(FXObject *, FXSelector, void *ptr)
{
  	if (game->engine.status == Status_Playing) {
		FXEvent *event = (FXEvent *)ptr;
		if (event->time - lastkeytime >= game_keytime) {
			game->engine.goRight();
			game->update();
			lastkeytime = event->time;
		}
	}
	return 1;
}

long MainWindow::onTimer(FXObject *, FXSelector, void *)
{
	if (game->engine.status == Status_EndLevel) {
		if (episode.level < episode.nolevels) {
  			FXMessageBox::information(this, MBOX_OK, "FX-Repton", "Level finished!  Press OK to play next level!\n(Remember to annotate the password!)");
			startLevel(episode, episode.level+1, false);
		} else {
  			FXMessageBox::information(this, MBOX_OK, "FX-Repton", "Episode finished!  Compliments!");
  			game->engine.status = Status_Idle;
    		}
	}
	if (game->engine.status == Status_TimeOut) {
		FXMessageBox::information(this, MBOX_OK, "FX-Repton", "Timeout, sorry!");
		game->engine.status = Status_Idle;
	}

	int delay;
	if (game->anim != Animation_None)
		delay = 20;
	else
		delay = 100;

	game->engine.timer(delay);
	game->update();

	if (game->engine.status == Status_Playing) {
		FXString s;
    		static int last_time = -1, last_diamonds = -1, last_score = -1;
    		static int last_level = -1, last_levelno = -1;
    		if (last_time != game->engine.game_time) {
			s.format(" %d ", game->engine.game_time / 1000);
			time_label->setText(s);
			last_time = game->engine.game_time;
		}
    		if (last_diamonds != game->engine.num_diamonds) {
			s.format(" %d ", game->engine.num_diamonds);
			diamonds_label->setText(s);
			last_diamonds = game->engine.num_diamonds;
		}
    		if (last_score != game->engine.score) {
			s.format(" %d ", game->engine.score);
			score_label->setText(s);
			last_score = game->engine.score;
		}
    		if (last_level != episode.level || last_levelno != episode.nolevels) {
			s.format(" %d/%d ", episode.level, episode.nolevels);
			level_label->setText(s);
			last_level = episode.level;
			last_levelno = episode.nolevels;
		}
  	}

	timerhandle = getApp()->addTimeout(delay, this, ID_TIMER);
	return 1;
}

