/*
 * Convert an original BBC Micro Repton map into a FX-Repton map.
 *
 * (c) 1996, 2002 Sandro Sigala
 *
 * Derived from the work done by Jasper Renow-Clarke (see notes.txt).
 */

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAP_WIDTH		28
#define MAP_HEIGHT		24
#define MAP_SIZE		(MAP_WIDTH * MAP_HEIGHT)

#define PASSWORD_OFFSET		0
#define TIME_LIMIT_OFFSET	64
#define EDITOR_CODES_OFFSET	80
#define TRANSPORTER_OFFSET	96
#define PALETTE_OFFSET		224
#define MAP_OFFSET		256
#define SPRITE_OFFSET		3616

enum object {
	/*
	 * Sprites.
	 */
	OBJ_ROCK = 0,
	OBJ_DIAMOND,
	OBJ_GROUND1,
	OBJ_GROUND2,
	OBJ_TIMECAPSULE,
	OBJ_SKULL,
	OBJ_EMPTY,
	OBJ_WALL,
	OBJ_WALL_WEST,
	OBJ_WALL_EAST,
	OBJ_WALL_NORTH,
	OBJ_WALL_SOUTH,
	OBJ_WALL_NORTH_WEST,
	OBJ_WALL_NORTH_EAST,
	OBJ_WALL_SOUTH_WEST,
	OBJ_WALL_SOUTH_EST,
	OBJ_FILLED_WALL,
	OBJ_FILLED_WALL_NORTH_WEST,
	OBJ_FILLED_WALL_NORTH_EAST,
	OBJ_FILLED_WALL_SOUTH_WEST,
	OBJ_FILLED_WALL_SOUTH_EST,
	OBJ_WALL_MIRROR,
	OBJ_STRONGBOX,
	OBJ_CAGE,
	OBJ_EGG,
	OBJ_KEY,
	OBJ_FUNGUS,
	OBJ_BOMB,
	OBJ_TRANSPORTER,
	OBJ_CROWN,
	OBJ_REPTON,
	OBJ_SPIRIT,
	OBJ_SPIRIT2,
	OBJ_MONSTER,
	OBJ_MONSTER2,
	OBJ_REPTON_LOOK_LEFT,
	OBJ_REPTON_LOOK_RIGHT,
	OBJ_REPTON_GO_RIGHT1,
	OBJ_REPTON_GO_RIGHT2,
	OBJ_REPTON_GO_RIGHT3,
	OBJ_REPTON_GO_RIGHT4,
	OBJ_REPTON_GO_LEFT1,
	OBJ_REPTON_GO_LEFT2,
	OBJ_REPTON_GO_LEFT3,
	OBJ_REPTON_GO_LEFT4,
	OBJ_REPTON_GO_UP1,
	OBJ_REPTON_GO_UP2,
	OBJ_BROKEN_EGG
};

void decode_pwd(unsigned char *buf)
{
	int i;
	for (i = 63; i >= 0; --i) {
		int c = *buf ^= i;
		if (c == 0x0d) c = '\0';
		*buf++ = tolower(c);
	}
}

void revbits(int *c, int n)
{
	int i, c1 = 0;
	for (i = 0; i < n; i++) {
		c1 = (c1 << 1) | (*c & 1);
		*c >>= 1;
	}
	*c = c1;
}

int count_transporters(unsigned char *map)
{
	int i, count = 0;
	for (i = 0; i < MAP_SIZE; ++i)
		if (map[i] == OBJ_TRANSPORTER)
			++count;
	return count;
}

int noerrors = 0;

void check_map(unsigned char *map, char transporters[16])
{
	int x, y;
	int n = count_transporters(map);
	int nspirit = 0, ncage = 0, nrepton = 0, nbomb = 0, nkey = 0, nstrongbox = 0;
	for (y = 0; y < MAP_HEIGHT; ++y)
		for (x = 0; x < MAP_WIDTH; ++x) {
			enum object o = map[x + y*MAP_WIDTH];
			if (o == OBJ_TRANSPORTER) {
				int i;
    				if (n == 0)
					fprintf(stderr, "Error: too much transporters\n"), ++noerrors;
				for (i = 0; i < 4; ++i)
					if (transporters[i*4] == x && transporters[i*4+1] == y)
						break;
				if (i == 4)
					fprintf(stderr, "Error: invalid transporter at %d, %d\n", x, y), ++noerrors;
    				--n;
   			} else if (o == OBJ_SPIRIT)
   				++nspirit;
			else if (o == OBJ_CAGE)
				++ncage;
			else if (o == OBJ_REPTON)
				++nrepton;
			else if (o == OBJ_BOMB)
				++nbomb;
			else if (o == OBJ_KEY)
				++nkey;
			else if (o == OBJ_STRONGBOX)
				++nstrongbox;
  		}

	if (nspirit != ncage)
		fprintf(stderr, "Error: %d spirits, %d cages\n", nspirit, ncage), ++noerrors;
	if (nrepton != 1)
		fprintf(stderr, "Error: %d reptons\n", nrepton), ++noerrors;
	if (nbomb != 1)
		fprintf(stderr, "Error: %d bombs\n", nbomb), ++noerrors;
	if (nstrongbox > 1 && nkey == 0)
		fprintf(stderr, "Error: %d strongboxes, zero keys\n", nstrongbox), ++noerrors;
}

void decode_map(unsigned char *map, char *buf)
{
	char *p = buf;
	int i, j, inbyte, outbyte, bits = 0;

	for (i = 0; i < MAP_SIZE; ++i) {
		outbyte = 0;
		for (j = 0; j < 5; ++j) {
			if (--bits <= 0) {
				inbyte = *p++;
				bits = 8;
			}
			outbyte = (outbyte << 1) | (inbyte & 1);
			inbyte >>= 1;
		}
		revbits(&outbyte, 5);
		map[i] = outbyte;
	}
}

/*
 * Prevent the (very) lazy user from finding the passwords.
 */
char *rot13(unsigned char *s)
{
	unsigned int i;
	unsigned char *p = s;
 	for (i = 0; i < strlen(s); ++i) {
		if (s[i] >= 'a' && s[i] <= 'z') {
 			s[i] = s[i] + 13;
 			if (s[i] > 'z')
 				s[i] = 'a' + s[i] - 'z' - 1;
		}
	}
	return p;
}

void xfread(void *buf, size_t size, size_t n, FILE *stream)
{
	if (fread(buf, size, n, stream) != n)
		fprintf(stderr, "Error: unexpected end of file\n"), ++noerrors;
}

void convert(FILE *fin, FILE *fout, int level)
{
	int ntrans, x, y;
	unsigned char map[MAP_SIZE];
	char buf[1024];
	int timelimit;
	unsigned char pwd[8][8];
	char password[9];
	char transporters[16];

	assert(level >= 0 && level <= 7);

	printf("Reading level %d\n\n", level + 1);
	printf("Reading password\n");
	fseek(fin, PASSWORD_OFFSET, SEEK_SET);
	fread(pwd, 1, sizeof(pwd), fin);
	decode_pwd(pwd[0]);
	strcpy(password, (char *)pwd[level]);
	password[8] = '\0';
	printf("  password: %s --> ", password);
	rot13(password);
	printf("%s\n", password);

	printf("Reading time\n");
	fseek(fin, TIME_LIMIT_OFFSET + level * 2, SEEK_SET);
	xfread(buf, 1, 2, fin);
	timelimit = (unsigned char)buf[0] + (unsigned char)buf[1] * 256;
	printf("  time: %d %d --> %d ", (unsigned char)buf[0], (unsigned char)buf[1], timelimit);
	timelimit = timelimit*60/1000;
	printf("= %d seconds\n", timelimit);

	printf("Reading transporters\n");
	fseek(fin, TRANSPORTER_OFFSET + level * 16, SEEK_SET);
	xfread(transporters, 1, 16, fin);

	printf("Reading map\n");
	fseek(fin, MAP_OFFSET + level*MAP_SIZE*5/8, SEEK_SET);
	xfread(buf, 1, MAP_SIZE*5/8, fin);
	decode_map(map, buf);

	printf("Checking map\n");
	check_map(map, transporters);

	ntrans = count_transporters(map);
	fprintf(fout, "%d %d %d %d %d\n", MAP_WIDTH, MAP_HEIGHT, ntrans, timelimit, 1);
	fprintf(fout, "%s\n", password);
	printf("  number of transporters: %d\n", ntrans);
	for (x = 0; x < ntrans; ++x) {
		printf("  %d %d %d %d\n", (int)transporters[x*4+0], (int)transporters[x*4+1],
			(int)transporters[x*4+2], (int)transporters[x*4+3]);
		fprintf(fout, "%d %d %d %d\n", (int)transporters[x*4+0], (int)transporters[x*4+1],
			(int)transporters[x*4+2], (int)transporters[x*4+3]);
		}
	for (y = 0; y < MAP_HEIGHT; ++y) {
		for (x = 0; x < MAP_WIDTH; ++x)
			fputc(map[x + y*MAP_WIDTH] + 'A', fout);
		fputc('\n', fout);
	}
	fprintf(fout, "-\n");

	printf("Done!\n");
}

int main(int argc, char **argv)
{
	int level;
	FILE *fin, *fout;

	if (argc != 4) {
		fprintf(stderr, "usage: filter <bbcmap> <destmap> <levelno>\n");
		return 1;
	}

	if ((fin = fopen(argv[1], "rb")) == NULL) {
		fprintf(stderr, "Error: cannot open %s for reading\n", argv[1]);
		return 1;
	}
 	if (argv[3][0] != '-') {
		if ((fout = fopen(argv[2], "wb")) == NULL) {
	  		fprintf(stderr, "Error: cannot open %s for writing\n", argv[2]);
  			return 1;
 		}

		level = argv[3][0] - '1';
		if (level < 0 || level > 7) {
  			fprintf(stderr, "Error: invalid level %s\n", argv[3]);
  			return 1;
 		}
	    	convert(fin, fout, level);
		fclose(fout);
  	} else {
  		char buf[128];
		for (level = 0; level < 8; ++level) {
			sprintf(buf, "%s%d", argv[2], level+1);
			if ((fout = fopen(buf, "wb")) == NULL) {
		  		fprintf(stderr, "Error: cannot open %s for writing\n", buf);
  				return 1;
 			}

		    	convert(fin, fout, level);
			fclose(fout);
  		}
   	}

	fclose(fin);

 	if (noerrors > 0)
 		fprintf(stderr, "\n>>> Number of errors: %d\n", noerrors);

	return 0;
}
